#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <mama_utils.h>
#include "astro.h"

/* @brief 
 *   Compute spherical distance
 *
 * @param[in] ra1
 *   ra from object 1
 * @param[in] dec1
 *   dec from object 1
 * @param[in] ra2
 *   ra from object 2
 * @param[in] dec2
 *   dec from object 2
 * @return
 *   spherical distance
 * compute the spherical distance between two object
*/
double distance( double ra1, double dec1, double ra2, double dec2)
{
  return acos( sin( dec1*DEG_TO_RAD)*sin( dec2*DEG_TO_RAD)+
    cos( dec1*DEG_TO_RAD)*cos( dec2*DEG_TO_RAD)* 
    cos( ra1*DEG_TO_RAD-ra2*DEG_TO_RAD))*RAD_TO_DEG;
}

/* @brief
 *   Compute ra dec from ksi and eta
 *   @param[in] ksi_deg in deg
 *   @param[in] eta_deg in deg
 *   @param[in] crval1 in deg
 *   @param[in] crval2 in deg
 *   @param[out] ra_deg in deg
 *   @param[out] dec_deg in deg
*/
void ksieta_to_radec(double ksi_deg, double eta_deg, double crval1, double crval2,
  double* ra_deg, double* dec_deg)
{
  double ksi_rad,eta_rad;
  double ra_off_rad,ra_rad,dec_rad;
  double ctan,ccos;

   ksi_rad = ksi_deg * DEG_TO_RAD;
   eta_rad = eta_deg * DEG_TO_RAD;

  ctan = tan( crval2*DEG_TO_RAD);
  ccos = cos( crval2*DEG_TO_RAD);

  ra_off_rad = atan2(ksi_rad/ccos,1.0-eta_rad*ctan);
  ra_rad = ra_off_rad + crval1*DEG_TO_RAD;
  
  if( ra_rad < 0.0){
    ra_rad = ra_rad + 2.0*M_PI;
  }
  else if ( ra_rad >=  (2.0*M_PI) ){
    ra_rad = ra_rad - 2.0*M_PI;
  }
  *ra_deg = ra_rad * RAD_TO_DEG;
  
  dec_rad = atan(cos(ra_off_rad)*((eta_rad+ctan)/(1.0-(eta_rad*ctan))));

  *dec_deg = dec_rad * RAD_TO_DEG; 

}

/*
 * @brief Convert x y pix to ra,dec using CD matrix
 *
 *  
*/
void cd_xypix_to_radec( double xpix, double ypix, double crval1, double crval2,
  double crpix1, double crpix2, double cd1_1, double cd1_2, double cd2_1,
  double cd2_2, double* ra_deg, double* dec_deg)
{
  double ksi_lin,eta_lin;

  ksi_lin = (xpix- crpix1)*cd1_1+(ypix-crpix2)*cd1_2;
  eta_lin = (xpix- crpix1)*cd2_1+(ypix-crpix2)*cd2_2;

  ksieta_to_radec( ksi_lin, eta_lin, crval1, crval2, ra_deg, dec_deg); 
}

/*
 * @brief
 *   convert alpha delta in pixel using cd matrix
 *
*/
void cd_radec_to_xypix( double crval1, double crval2, double crpix1, 
  double crpix2, double cd1_1, double cd1_2, double cd2_1, 
  double cd2_2, double ra_deg, double dec_deg, double* xpix,
  double* ypix)
{
  
  double dr;
  double raobj_rad,decobj_rad,raref_rad,decref_rad;
  double ksi_deg,eta_deg;  


  if( (cd1_2!=0.0) || (cd2_1!=0.0)){
    printf("The CD MATRIX must be diagonal\n");
    exit(-1);
  }

  raobj_rad = ra_deg * DEG_TO_RAD;
  decobj_rad = dec_deg*DEG_TO_RAD;
  raref_rad = crval1 * DEG_TO_RAD;
  decref_rad = crval2 * DEG_TO_RAD; 

  dr = sin(decobj_rad)*sin(decref_rad)+cos(decobj_rad)*cos(decref_rad)*
    cos(raobj_rad-raref_rad);
  
  ksi_deg = cos(decobj_rad)*sin(raobj_rad-raref_rad)/dr*RAD_TO_DEG;
  eta_deg = (sin(decobj_rad)*cos(decref_rad)-cos(decobj_rad)*sin(decref_rad)*
      cos(raobj_rad-raref_rad))/dr*RAD_TO_DEG;
  *xpix = crpix1 + ksi_deg/cd1_1;
  *ypix = crpix2 + eta_deg/cd2_2;
}


#define AMD( RES, X, Y, ORDER, COEF, k)\
  for(i=ORDER; i>=0; i--){\
    RES += COEF[k]*pow(X,i)*pow(Y,ORDER-i);\
    k++;\
  }
/*
 * @brief Convert x y pixels to ra,dec using AMD coefficients x pixel 1 => [0.5 1.5] center at 1,1

  for now only order 3 and type 1 supported !!
*/
void amd_xypix_to_radec( double xpix, double ypix, amd_wcs* wcs,
  double* ra_deg, double* dec_deg)
{
  double xmm, ymm,  xmm2, ymm2, xmm3, ymm3, x2y2;
  double ksi_amd, eta_amd;
  int order,i,k,l = 0;
  double x,y;
  
  /* Convert from image pixels to plate pixels */
  x = xpix + wcs->xpixel_offset - 1.0 + 0.5;
  y = ypix + wcs->ypixel_offset - 1.0 + 0.5;
  /* Convert from pixels to millimeters */
  xmm = ( wcs->ppo[2]  - x * wcs->xpix_size )/1000.0 ;
  ymm = (y * wcs->ypix_size  - wcs->ppo[5] ) /1000.0 ;
  
  xmm2 = xmm*xmm;
  ymm2 = ymm*ymm;
  xmm3 = xmm*xmm2;
  ymm3 = ymm*ymm2;
  x2y2 = xmm2+ymm2;

  ksi_amd = wcs->amdx[0]*xmm + wcs->amdx[1]*ymm + wcs->amdx[2] + wcs->amdx[3]*xmm2 + 
    wcs->amdx[4]*xmm*ymm + wcs->amdx[5]*ymm2 + wcs->amdx[ 6]*(x2y2) +
    wcs->amdx[7]*xmm3 + wcs->amdx[8]*xmm2*ymm + wcs->amdx[9]*xmm*ymm2 +
    wcs->amdx[10]*ymm3 + wcs->amdx[11]*xmm*(x2y2) + wcs->amdx[12]*xmm*x2y2*x2y2;

  eta_amd = wcs->amdy[0]*ymm + wcs->amdy[1]*xmm + wcs->amdy[2] + wcs->amdy[3]*ymm2 + 
   wcs->amdy[4] * xmm * ymm + wcs->amdy[5] * xmm2 + wcs->amdy[ 6]*(x2y2) +
   wcs->amdy[7] * ymm3 + wcs->amdy[8] * ymm2*xmm + wcs->amdy[9] *ymm*xmm2 +
   wcs->amdy[10]*xmm3 + wcs->amdy[11]*ymm*(x2y2) + wcs->amdy[12]*ymm*x2y2*x2y2;

  if (wcs->reduction_order == 6){
    k = 13;
    l = 13;
    for(order=4; order<=6; order++){
      AMD(ksi_amd, xmm, ymm, order,wcs->amdx,k);
      AMD(eta_amd, ymm, xmm, order,wcs->amdy,l);
      l += order-1;
      k += order-1;
    }
  }
  ksi_amd /= 3600.0;
  eta_amd /= 3600.0;
  ksieta_to_radec( ksi_amd, eta_amd, wcs->ra_deg, wcs->dec_deg, ra_deg, dec_deg);
}

/* 
 * @brief Convert x y pixels in  alpha delta using PV
 *
*/

#define PV( RES, X, Y, ORDER, COEF, k)\
  for(i=ORDER; i>=0; i--){\
    double x,y;\
    x = pow(X,i);\
    y = pow(Y,ORDER-i);\
    RES += COEF[k]*x*y;\
    k++;\
    if ( (k == 3) || (k == 11) || (k == 23)){\
      k++;\
    }\
  }


/* x pixel 1 => [0.5 1.5] center at 1,1 */

void pv_xypix_to_radec( double xpix, double ypix, pv_wcs* wcs,
  double* ra_deg, double* dec_deg)
{
  double ksi_lin,eta_lin;
  int order = 6, k = 0;
  double ksi_pv = 0.0, eta_pv = 0.0;
  int i,j;
  
  ksi_lin = (xpix- wcs->crpix1)*wcs->cd1_1+(ypix-wcs->crpix2)*wcs->cd1_2;  
  eta_lin = (xpix- wcs->crpix1)*wcs->cd2_1+(ypix-wcs->crpix2)*wcs->cd2_2;

  /*printf("ksilin %.10E etalin %.10E\n", ksi_lin,eta_lin);*/
  /* pv */
  k = 0;
  for(j=0; j<=order; j++){
    PV( ksi_pv, ksi_lin, eta_lin, j, wcs->pv1, k);
  }
  k = 0;
  for(j=0; j<=order; j++){
    PV( eta_pv, eta_lin, ksi_lin, j, wcs->pv2, k);
    /*printf("ORDER %d ksi_pv %.10E\n",j,eta_pv);*/
  }
  
  /*printf("PV ksi_pv %.10E eta_pv %.10E\n", ksi_pv,eta_pv); */

  ksieta_to_radec( ksi_pv, eta_pv, wcs->crval1, wcs->crval2,
    ra_deg, dec_deg);
}

/* @brief
 *   Load an amd result structure
 *
*/
amd_wcs* amd_wcs_load(const char* filename)
{
  FILE* file = NULL;
  amd_wcs* result = NULL;
  double rah, ram, ras;
  double decd, decm, decs;
  char* sign = NULL;
  char record[90];
  int i;
  int coef = 0;

  if(filename != NULL){
    file = fopen(filename,"r");
    if (file == NULL) {
      printf("impossible d'ouvir le fichier %s\n",filename);
      exit(-1);
    }

    result = (amd_wcs*) (calloc(sizeof(amd_wcs),1));
    
    result->xpixel_offset = result->ypixel_offset = 0.0;
    if( result == NULL ){
      printf("Impossible d'allouer de la memoire pour le result");
      exit(-1);
    }

    while(fgets(record,90,file)){
      char* key = strtok(record, "=");
      char* value = strtok(NULL,"/\n");

      if ( strstr(key, "REDUORD") != NULL){
        result->reduction_order = atoi(value);
	for (i=result->reduction_order-1; i >=1; i--){
	  coef += i;
	}
	for( i=1; i<= result->reduction_order+1; i++){
          coef += i;
	}	
      }
      else if ( strstr(key, "PLTRAH") != NULL){
        rah = atof(value);
      }
      else if ( strstr(key, "PLTRAM") != NULL){
        ram = atof(value);
      }
      else if ( strstr(key, "PLTRAS") != NULL){
        ras = atof(value);
      }
      else if ( strstr(key, "PLTDECSN") != NULL){
        sign = strstr(value, "-");
      }
      else if ( strstr(key, "PLTDECD") != NULL){
        decd = atof(value);
      }
      else if ( strstr(key, "PLTDECM") != NULL){
        decm = atof(value);
      }
      else if ( strstr(key, "PLTDECS") != NULL){
        decs = atof(value);
      }
      else if ( strstr(key, "XPIXELSZ") != NULL){
        result->xpix_size = atof(value);
      }
      else if ( strstr(key, "YPIXELSZ") != NULL){
        result->ypix_size = atof(value);
      }
      else if ( strstr(key, "CNPIX1") != NULL){
	result->xpixel_offset = atof(value);
      }
      else if ( strstr(key, "CNPIX2") != NULL){
	result->ypixel_offset = atof(value);
      }
      else if ( strstr(key, "PPO") != NULL){
        for(i=1; i<=6;i++){
	  char ppo[6];
	  sprintf(ppo,"PPO%d",i);
          if (strstr(key, ppo) != NULL){
	    result->ppo[i-1] = atof(value);
	    break;
	  }
	}
      }
      else if ( strstr(key, "AMDX") != NULL){
        for(i=1; i<=coef;i++){
	  char amdx[7];
	  double v;
	  sprintf(amdx,"AMDX%d ",i);
          if (strstr(key, amdx) != NULL){
	    v = atof(value);
	    result->amdx[i-1] = v;
	    break;
	  }
	}
      }
      else if ( strstr(key, "AMDY") != NULL){
        for(i=1; i<=coef;i++){
	  char amdy[7];
	  sprintf(amdy,"AMDY%d ",i);
          if (strstr(key, amdy) != NULL){
	    result->amdy[i-1] = atof(value);
	    break;
	  }
	}
      }
    }
    result->ra_deg = (rah+ram/60.0+ras/3600.0)*15;
    result->dec_deg = decd+decm/60.0+decs/3600.0;

/*    for(i=0; i< coef; i++){
      printf("coef %d = %E\n",i,result->amdx[i]);
    }
*/
    if (sign != NULL) {
      result->dec_deg = -result->dec_deg;
    }
    fclose(file);
  }
  return result;
}

/* 
 * @brief amd result free
 *
*/
void amd_result_free(amd_wcs** wcs)
{
  if (wcs && *wcs){
    free(wcs);
    *wcs = NULL;
  }
}



/* @brief
 *   Load a pw result structure
 *
*/
pv_wcs* pv_wcs_load(const char* filename)
{
  FILE* file = NULL;
  pv_wcs* result = NULL;
  char record[90];
  int i;

  MAMA_NOT_NULL( filename);

  MAMA_FOPEN( file, filename, "r");

  MAMA_CALLOC( result, pv_wcs, 1);

    while(fgets(record,90,file)){
      char* key = strtok(record, "=");
      char* value = strtok(NULL,"/\n");

      if ( strstr(key, "CRVAL1") != NULL){
        result->crval1 = atof(value);
      }
      else if ( strstr(key, "CRVAL2") != NULL){
        result->crval2 = atof(value);
      }
      else if ( strstr(key, "CRPIX1") != NULL){
        result->crpix1 = atof(value);
      }
      else if ( strstr(key, "CRPIX2") != NULL){
        result->crpix2 = atof(value);
      }
      else if ( strstr(key, "CD1_1") != NULL){
        result->cd1_1 = atof(value);
      }
      else if ( strstr(key, "CD1_2") != NULL){
        result->cd1_2 = atof(value);
      }
      else if ( strstr(key, "CD2_1") != NULL){
        result->cd2_1 = atof(value);
      }
      else if ( strstr(key, "CD2_2") != NULL){
        result->cd2_2 = atof(value);
      }
      else if ( strstr(key, "PV1") != NULL){
        for(i=0; i<=30;i++){
	  char pv[8];
	  sprintf(pv,"PV1_%d ",i);
          if (strstr(key, pv) != NULL){
	    result->pv1[i] = atof(value);
	    break;
	  }
	}
      } 
      else if ( strstr(key, "PV2") != NULL){
        for(i=0; i<=30;i++){
	  char pv[8];
	  sprintf(pv,"PV2_%d ",i);
          if (strstr(key, pv) != NULL){
	    result->pv2[i] = atof(value);
	    break;
	  }
	}
      }
    }
    fclose(file);

  return result;
}
