#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <cutout.h>
#include <math.h>
#include <getopt.h>
#include <astro.h>

/*
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* DEFINITIONS
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/

#define         CUTOUT_SYNTAX \
"cutout                      \n\
                -params   <display cutout params> \n\
		-get      <create a cut file>     \n"

#define         CUTOUT_PARAMS_SYNTAX \
"cutout                         \n\
                -file   <file to cut>\n\
		-center <x y>\n\
		-box    <x y>\n\
Example :\n\
  cutout -params -file mymosaique.fits -center 10 10 -box 10 10"

#define         CUTOUT_GET_SYNTAX \
"cutout                         \n\
                -file   <file to cut>\n\
		-naxis  <x y>\n\
		-crpix  <x y>\n\
Example :\n\
  cutout -params -file mymosaique.fits -center 10 10 -box 10 10"

#define PI 3.1415926535897932384626433832795028841971693993751058209749445923078164

/*
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* main program
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/
int main(int argc,char* argv[]){

  if(argc < 2){
    printf(CUTOUT_SYNTAX);
    exit(-1);
  }

  
  // cutout mode 
  if( strcmp("-params",argv[1])==0 ){
    cutout_params(argc,argv);  
  }
  else if(strcmp("-get",argv[1])==0 ){
    cutout_get(argc,argv);
  }
  else{
    printf(CUTOUT_SYNTAX);
    exit(-1);
  }
}

/*
----------------------------------------------------------------------------------------
*/

/*
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* cutout_params
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/
int cutout_params(int argc,char* argv[]){
  int i;
  char* filename = NULL;
  double ra_deg,dec_deg,ramin,ramax,decmin,decmax;
  double alpha_size,delta_size;
  fits* result = NULL;
  char value[22];
  unsigned long firstx,firsty;
  double crpix1,crpix2,cd1_1,cd1_2,cd2_1,cd2_2,crval1,crval2;
  double x_center_pix,y_center_pix;
  double x_left_pix,x_right_pix;
  double y_bottom_pix,y_top_pix;
  long box_width,box_height;
  unsigned int width,height;

  if(argc < 10){
    printf(CUTOUT_PARAMS_SYNTAX);
    exit(-1);
  }
  for( i=2;i<argc;i++ ){
    if( strcmp("-file",argv[i])==0 ){
      filename = strdup(argv[++i]);  
    }
    else if(strcmp("-center",argv[i])==0 ){
      ra_deg = atof(argv[++i]);
      dec_deg = atof(argv[++i]);
    }
    else if(strcmp("-box",argv[i])==0 ){
      alpha_size = atof(argv[++i])/2.0;
      delta_size = atof(argv[++i])/2.0;

    }    
    else{
      printf(CUTOUT_PARAMS_SYNTAX);
      exit(-1);
    }
  }

  /* analyse the parameters */
  if(filename == NULL){
    printf("You must provided a file name with option -file\n");
    exit(-1);
  }
  if (((dec_deg+delta_size) > 90) ||((dec_deg-delta_size) < -90)){
    printf("You must provided good parameter for DEC center and delta box\n");
    exit(-1);
  }
  
  if(floor(dec_deg)<90)
  if(((ra_deg+alpha_size/2.0) > 360) ||((ra_deg-alpha_size/2.0) < 0)){
    printf("You must provided good parameter for RA center and alpha box\n");
    exit(-1);
    }

  /* computes the cutout parameters */ 
  result = fits_new(filename);
  
  /* read the reference system */
  if(fits_keyword_get(result,"CRPIX1",value)){
    crpix1 = atof(value);
  }
  else {
    printf("Can't read CRPIX1 in header\n");
    exit(-1);
  }
  if(fits_keyword_get(result,"CRPIX2",value)){
    crpix2 = atof(value);
  }
  else {
    printf("Can't read CRPIX2 in header\n");
    exit(-1);
  }
  if(fits_keyword_get(result,"CRVAL1",value)){
    crval1 = atof(value);
  }
  else {
    printf("Can't read CRVAL1 in header\n");
    exit(-1);
  }
  if(fits_keyword_get(result,"CRVAL2",value)){
    crval2 = atof(value);
  }
  else {
    printf("Can't read CRVAL2 in header\n");
    exit(-1);
  }
  if(fits_keyword_get(result,"CD1_1",value)){
    cd1_1 = atof(value);
  }
  else {
    printf("Can't read CD1_1 in header\n");
    exit(-1);
  }
  if(fits_keyword_get(result,"CD1_2",value)){
    cd1_2 = atof(value);
  }
  else {
    printf("Can't read CD1_2 in header\n");
    exit(-1);
  }
  if(fits_keyword_get(result,"CD2_1",value)){
    cd2_1 = atof(value);
  }
  else {
    printf("Can't read CD2_1 in header\n");
    exit(-1);
  }
  if(fits_keyword_get(result,"CD2_2",value)){
    cd2_2 = atof(value);
  }
  else {
    printf("Can't read CD2_2 in header\n");
    exit(-1);
  }
  /* computes in naxis2 */
  cd_radec_to_xypix(crval1,crval2,crpix1,crpix2,cd1_1,cd1_2,cd2_1,
    cd2_2,ra_deg,dec_deg-delta_size/2.0,&x_left_pix,&y_bottom_pix);
  cd_radec_to_xypix(crval1,crval2,crpix1,crpix2,cd1_1,cd1_2,cd2_1,
    cd2_2,ra_deg,dec_deg+delta_size/2.0,&x_left_pix,&y_top_pix);

  if(y_top_pix >= result->naxis2){
    y_top_pix = result->naxis2-1;
  }
  if(y_bottom_pix < 0){
    y_bottom_pix = 0;
  }
     
  height = lrint(y_top_pix-y_bottom_pix+1);
  
  /* computes in naxis1 */
  cd_radec_to_xypix(crval1,crval2,crpix1,crpix2,cd1_1,cd1_2,cd2_1,
    cd2_2,(alpha_size/2.0)/cos(dec_deg/180.0*PI)+ra_deg,dec_deg,
    &x_left_pix,&y_top_pix);

  cd_radec_to_xypix(crval1,crval2,crpix1,crpix2,cd1_1,cd1_2,cd2_1,
    cd2_2,ra_deg-(alpha_size/2.0)/cos(dec_deg/180.0*PI),dec_deg,
    &x_right_pix,&y_top_pix);

 if(x_right_pix >= result->naxis1){
    x_right_pix = result->naxis2-1;
  }
  if(x_left_pix < 0){
    x_left_pix = 0;
  }
  width = (long)rint(x_right_pix-x_left_pix+1);

  /* transform in pixels */
  firstx = (long)rint(x_left_pix);
  firsty = (long)rint(y_bottom_pix);
    
  printf("first %ld,%ld box %u,%u\n",firstx,firsty,width,height);

  /* compute the new position of reference pixel */
  crpix1 = crpix1-firstx;
  crpix2 = crpix2-firsty;

  /* find the center */
  cd_xypix_to_radec(width/2.0,height/2.0,crval1,crval2,crpix1,crpix2,cd1_1,
		 cd1_2,cd2_1,cd2_2,&ra_deg,&dec_deg);

  printf("TESTS\n");
  printf("RA_C %.13f\n",ra_deg); 
  printf("DEC_C %.13f\n",dec_deg);
  printf("NAXIS %u %u\n",width,height);
  printf("CRPIX %.13f %.13f\n",crpix1,crpix2);
  printf("CRVAL %.13f %.13f\n",crval1,crval2);
  printf("CD %.13f %.13f %.13f %.13f\n",cd1_1,cd1_2,cd2_1,cd2_2);
}

/*
----------------------------------------------------------------------------------------
*/

/*
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* cutout_get
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/
int cutout_get(int argc,char* argv[]){

  int i,nbread;
  double oldcrpix1,oldcrpix2,crpix1,crpix2;
  long naxis1,naxis2;
  fits* result = NULL;
  char* filename = NULL;
  long firstx,firsty;
  char* line = NULL;
  long size_pixel,size_to_complete;
  char value[70];
  int jpg = 0;

  if(argc < 10){
    printf(CUTOUT_GET_SYNTAX);
    exit(-1);
  }
  for( i=2;i<argc;i++ ){
    if( strcmp("-file",argv[i])==0 ){
      filename = strdup(argv[++i]);  
    }
    else if(strcmp("-naxis",argv[i])==0 ){
      naxis1 = atol(argv[++i]);
      naxis2 = atol(argv[++i]);
    }
    else if(strcmp("-crpix",argv[i])==0 ){
      crpix1 = atof(argv[++i]);
      crpix2 = atof(argv[++i]);

    }
   else if(strcmp("-jpg",argv[i])==0 ){
     jpg = 1;
   }
   else{
      printf(CUTOUT_GET_SYNTAX);
      exit(-1);
    }
  }
  // analyse the parameters
  if(filename == NULL){
    printf("You must provided a file name with option -file\n");
    exit(-1);
  }

  // read reference pixel
  result = fits_new(filename);

  if(fits_keyword_get(result,"CRPIX1",value)){
    oldcrpix1 = atof(value);
  }
  else {
    printf("Can't read CRPIX1 in header\n");
    exit(-1);
  }


  if(fits_keyword_get(result,"CRPIX2",value)){
    oldcrpix2 = atof(value);
  }
  else {
    printf("Can't read CRPIX2 in header\n");
    exit(-1);
  }

  firstx = oldcrpix1 - crpix1;
  firsty = oldcrpix2 - crpix2;

  // write th new position
  sprintf(value,"%20E",crpix1);
  fits_keyword_set(result,"CRPIX1",value);
  sprintf(value,"%20E",crpix2);
  fits_keyword_set(result,"CRPIX2",value);  

  // write new naxis
  sprintf(value,"%20ld",naxis1);
  fits_keyword_set(result,"NAXIS1",value);
  sprintf(value,"%20ld",naxis2);
  fits_keyword_set(result,"NAXIS2",value);

  // transform for JPG
  if(jpg == 1){
   sprintf(value,"%20d",32768);
   fits_keyword_set(result,"BZERO",value);
   sprintf(value,"%20d",1);
   fits_keyword_set(result,"BSCALE",value);
  }
  // write in the cut file
  nbread = fwrite(result->header, 2880 * result->nb_header_block,1,stdout);

  if (nbread != 1) {
    error(-1,"Can't write the header in the cutout file");
  }
  
  if(result->bitpix == 16){
    size_pixel = 2;
  }
  if(result->bitpix == -32){
    size_pixel = 4;
  }

  line = (char*)malloc(size_pixel*naxis1);
  
  /* doing the cutout */
  
  /* go to the first line */
  fseek(result->file,result->nb_header_block*2880+firstx*size_pixel+firsty*size_pixel*result->naxis1,SEEK_SET);
  /* read the image */
  for(i=0; i < naxis2;i++){
    nbread = fread(line,naxis1,size_pixel,result->file);
    nbread = fwrite(line,naxis1,size_pixel,stdout);
    fseek(result->file,(result->naxis1-naxis1)*size_pixel,SEEK_CUR);
  }
  free(line);
  result->naxis1 = naxis1;
  result->naxis2 = naxis2;
  fits_complete_with_zero(result,stdout);

  fits_free(result);
}
/*
----------------------------------------------------------------------------------------
*/
#undef PI
